You can search for identifiers within the package.
in-package search v0.2.0
sek
Sek.CAPACITY
val capacity : int -> int
The function capacity receives a nonnegative depth k and must return the desired capacity of a chunk at depth k. This capacity must be at least 2.
capacity
k
2